require "enumerator"
require "forwardable"
require "imlib2"
require "weakref"


class Canvas
    
    extend(Forwardable)
    DEFAULT_WIDTH= 800
    DEFAULT_HEIGHT= 600
    @@instances= {}
    
    def self.get(id)
      begin
        return @@instances[id].__getobj__ if @@instances[id]
      rescue WeakRef::RefError
      end
      return Canvas.new(id)
    end
    
    def initialize(id, create= false)
      raise("Invalid id: #{id}") if !(id=~/\A[A-Za-z0-9]+\z/)
      @id= id
      @path= "canvas/#{id}.png"
      @image_data= nil
      @last_update= Time.now()
      @color= Imlib2::Color::BLACK
      @mutex= Mutex.new()
      @cv= ConditionVariable.new()
      if File.exist?(@path)
        @image= Imlib2::Image.load(@path)
      elsif create
        @image= Imlib2::Image.new(DEFAULT_WIDTH, DEFAULT_HEIGHT)
        @image.fill_rect([0, 0, DEFAULT_WIDTH, DEFAULT_HEIGHT], Imlib2::Color::WHITE)
        save_image()
      else
        raise("No such id")
      end
      @@instances[id]= WeakRef.new(self)
    end
    
    attr_reader(:image, :last_update, :id)
    def_delegators(:@mutex, :synchronize)
    def_delegators(:@image, :width, :height)
    
    def save_image()
      @image.save(@path)
      @image_data= nil
      @last_update= Time.now()
      @cv.broadcast()
    end
    
    def image_data()
      return @image_data||= open(@path, "rb"){ |f| f.read() }
    end
    
    def wait_for_change(since, timeout)
      begin
        Timeout.timeout(timeout) do
          while @last_update <= since
            @cv.wait(@mutex)
          end
        end
        return true
      rescue TimeoutError
        return false
      end
    end
    
    def draw_line(from, to)
      #imlib2が上方向の線をうまく引けない（途切れる）ようなので、
      #必ず下向きの線になるようにする。
      args= [from, to].sort_by(){ |n| n[1] } + [@color]
      @image.draw_line(*args)
    end
    
    def draw_line_series(nodes)
      nodes.each_cons(2) do |from, to|
        draw_line(from, to)
      end
    end
    
    def draw_dots(nodes, radius)
      for node in nodes
        @image.fill_ellipse(node, [radius, radius], @color)
      end
    end
    
    def erase(from, to)
      @image.fill_rect(pos_to_range(from, to), Imlib2::Color::WHITE)
    end
    
    def pos_to_range(from, to)
      return [
        [from[0], to[0]].min,
        [from[1], to[1]].min,
        (from[0]-to[0]).abs,
        (from[1]-to[1]).abs,
      ]
    end
    
    def color=(c)
      rgba= c.split(/,/).map(){ |s| s.to_i() } + [255]
      @color= Imlib2::Color::RgbaColor.new(rgba)
    end

end
