require "enumerator"
require "ostruct"
require "webrick"
require "json"
require "gimite_ext"
require "canvas"


class Session < OpenStruct
    
    @@instances= {}
    @@mutex= Mutex.new()
    
    def self.get(req, res, allow_create= true)
      session_cookie= req.cookies.find(){ |c| c.name=="session" }
      session_cookie||= WEBrick::Cookie.new("session", rand(2**30).to_s()) if allow_create
      return nil if !session_cookie
      res.cookies.push(session_cookie)
      id= session_cookie.value
      @@mutex.synchronize() do
        return @@instances[id]||= Session.new({:session_id => id})
      end
    end
    
end


class Localization
    
    @@instances= {}
    
    def self.get(lang)
      #return @@instances[lang]||= Localization.new(lang)
      return Localization.new(lang) #とりあえずデバッグ用にキャッシュしない。
    end
    
    def initialize(lang)
      raise("Invalid language") if !(lang=~/\A[a-z]+\z/)
      @lang= lang
      @table= IO.enum_for(:foreach, "local/#{lang}").
        map_to_hash(){ |s| s.chomp().split(/\s+/, 2) }
    end
    
    def method_missing(name)
      if @table[name.to_s()]
        return @table[name.to_s()]
      else
        raise(NameError, "Missing '#{@lang}' localization for '#{name}'")
      end
    end
    
end


class Controller
    
    MAX_LONG_LIFETIME= 90 #長寿コネクションの寿命の最大値。
    MIN_LONG_LIFETIME= 0 #長寿コネクションの寿命の最小値。すぐに更新があってもこの秒数は待つ。
    LONG_TIMEOUT= 100 #長寿コネクションがこの時間帰ってこないと、切れたと判断。
    DRAW_TIMEOUT= 60 #描画コマンド送信後この時間応答がないと、切れたと判断。
    RETRY_FREQ= 20 #再接続の試行頻度。
    COMMAND_BUFFER_TIME= 2 #この時間内のコマンドはまとめて送信。
    
    def initialize(root_url, port)
      @root_url= root_url.gsub(/\/$/, "")
      @server= WEBrick::HTTPServer.new(
        :Port => port,
        :MimeTypes => WEBrick::HTTPUtils.load_mime_types("config/mime.types")
      )
      @server.mount_proc("/", method(:route))
      for name in ["js", "img"]
        @server.mount("/#{name}", WEBrick::HTTPServlet::FileHandler, name)
      end
    end
    
    def run()
      trap("INT"){ Process.kill(:TERM, Process.pid) }
      @server.start()
    end
    
    def route(req, res)
      (dummy, canvas_id, op)= req.path.split(/\//)
      if !canvas_id
        top_page(req, res)
      elsif canvas_id=="create"
        create(req, res)
      else
        canvas= Canvas.get(canvas_id)
        case op
          when nil
            if req.path=~/\/$/
              index(canvas, req, res)
            else
              redirect(res, canvas_base_url(canvas)+"/")
            end
          when "image"
            image(canvas, req, res)
          when "command"
            command(canvas, req, res)
          else
            raise("Unknown operation")
        end
      end
    end

    def top_page(req, res)
      local= localization(req)
      res["Content-Type"]= "text/html"
      res.body= render("top_page", binding())
    end

    def create(req, res)
      raise("must be POST") if req.request_method!="POST"
      canvas_id= (0...10).map(){ (?a+rand(26)).chr }.join("")
      canvas= Canvas.new(canvas_id, true)
      redirect(res, canvas_base_url(canvas)+"/")
    end

    def index(canvas, req, res)
      local= localization(req)
      root_url= @root_url
      canvas_id= canvas.id
      base_url= canvas_base_url(canvas)
      width= canvas.width
      height= canvas.height
      long_timeout= LONG_TIMEOUT
      draw_timeout= DRAW_TIMEOUT
      retry_freq= RETRY_FREQ
      command_buffer_time= COMMAND_BUFFER_TIME
      res["Content-Type"]= "text/html"
      res.body= render("index", binding())
    end

    def command(canvas, req, res)
      canvas.synchronize() do
        for command in JSON.parse(req.body)
          p command
          canvas.color= command["color"] if command["color"]
          case command["mode"]
            when "line"
              canvas.draw_line(command["from"], command["to"])
            when "line_series"
              canvas.draw_line_series(command["nodes"])
            when "dots"
              canvas.draw_dots(command["nodes"], command["radius"])
            when "erase"
              canvas.erase(command["from"], command["to"])
          end
        end
        canvas.save_image()
        res["Content-Type"]= "text/html"
        res.body= {"revision" => canvas.last_update.to_f()}.to_json()
      end
    end

    def image(canvas, req, res)
      block= req.query["block"]
      save= req.query["save"]
      session= Session.get(req, res, !block)
        #Cookieが無効な環境でリクエストを連発するのを防ぐため、
        #セッションがない状態でのblockモードは許可しない。
      if !session
        res.status= 400
        return
      end
      sleep(MIN_LONG_LIFETIME) if block
      canvas.synchronize() do
        if block && session.last_send
          canvas.wait_for_change(session.last_send, MAX_LONG_LIFETIME-MIN_LONG_LIFETIME)
        end
        session.last_send= canvas.last_update
        res["Content-Type"]= "image/png"
        res["Content-Disposition"]= "attachment; filename=sharpan-#{canvas.id}.png" if save
        res.body= canvas.image_data
        res.cookies.push(WEBrick::Cookie.new("#{canvas.id}/revision", session.last_send.to_f.to_s))
      end
    end
    
    def render(template_name, bind)
      return ERB.new(IO.read("view/#{template_name}.rhtml")).result(bind)
    end
    
    def canvas_base_url(canvas)
      return "#{@root_url}/#{canvas.id}"
    end
    
    def redirect(res, url)
      res.status= 301
      res["Location"]= url
    end
    
    def localization(req)
      req["accept-language"]=~/ja|en/
      lang= $&=="ja" ? "ja" : "en" #Todo: アドホック。
      return Localization.get(lang)
    end
    
end
