module Kernel
    
  private
    
    #引数を返すようにしてみる。[ruby-dev:29736]
    def p(*args)
      args.each(){ |a| puts(a.inspect) }
      return args.last
    end
    
    def assert(cond)
      raise("Assertion failed") if !cond
    end
    
end


class Object
    
    #このオブジェクトの特異クラス。
    def singleton_class
      class << self
        self
      end
    end
    
    #特異メソッドを定義する。
    def define_singleton_method(*args, &block)
      singleton_class.instance_eval() do
        define_method(*args, &block)
      end
    end
    
    #ブロック付き特異メソッドを定義する。
    #ブロックは最後の引数として渡される。
    def define_singleton_method_with_block(*args, &block)
      singleton_class.define_method_with_block(*args, &block)
    end
    
end


class Module
    
    #member(メソッドorインスタンス変数)が等しければオブジェクト同士が等価になるように、
    #==, eql?, hashを定義する。
    def def_equal_by(member)
      if member.to_s()=~/^@/
        getter= proc(){ |o| o.instance_variable_get(member) }
      else
        getter= proc(){ |o| o.__send__(member) }
      end
      define_method(:==) do |rhs|
        self.class==rhs.class && getter.call(self)==getter.call(rhs)
      end
      alias_method(:eql?, :==)
      define_method(:hash){ getter.call(self).hash }
    end
    
    #クラスメソッドをインスタンスメソッドとして呼べるようにする。
    def static_method(*method_names)
      if !method_names.empty?
        method_names.each() do |name|
          define_method(name) do |*args|
            self.class.__send__(name, *args){ |*bas| yield(*bas) }
          end
        end
      else
        raise(ArgumentError, "method list is empty")
      end
    end
    
    #指定のメソッドの戻り値を、同名のインスタンス変数にキャッシュする。
    def cache_return_value(*method_names)
      method_names.each() do |name|
        orig_method= instance_method(name)
        if orig_method.arity!=0
          raise(ArgumentError, "cache_return_value don't support methods with arity!=0")
        end
        iv_name= "@#{name}"
        define_method(name) do
          if (result= instance_variable_get(iv_name)) == nil
            instance_variable_set(iv_name, orig_method.bind(self).call())
          else
            result
          end
        end
      end
    end
    
    #ブロック付きメソッドを定義する。
    #ブロックは最後の引数として渡される。
    #UtilRb (http://www.laas.fr/~sjoyeux/research.php) より。
    def define_method_with_block(name, &mdef)
      class_eval(
        "def #{name}(*args, &block)\nargs << block\ndmwb_#{name}_user_definition(*args)\nend\n")
      define_method("dmwb_#{name}_user_definition", &mdef)
    end
    
end


class Exception
    
    def print_backtrace(io= $stderr)
      io.printf("%s: %s (%p)\n", self.backtrace[0], self.message, self.class)
      for s in self.backtrace[1..-1]
        io.printf("        %s\n", s)
      end
    end
    
end


module Enumerable
    
    #各要素についてblockを呼び出し、その合計を返す。
    #要素が無ければzeroを返す。
    def sigma(zero= 0, &block)
      sum= nil
      block||= proc(){ |v| v }
      for v in self
        sum= sum ? (sum+block.call(v)) : block.call(v)
      end
      return sum || zero
    end
    
    #mapのHash版。各要素をblockを使って変換し、
    #その結果([key, value])を新しいHashにして返す。
    def map_to_hash(h= {}, &block)
      for a in self
        (k, v)= block ? yield(a) : a
        h[k]= v
      end
      return h
    end
    
    #各要素をblockを使って変換した結果をキーとしてグループ分けする。
    #e.g. [1, 2, 3, 4, 5].group_by(){ |a| a%2 }
    #     #==> {0=>[2, 4], 1=>[1, 3, 5]}
    def group_by(&block)
      h= {}
      for a in self
        k= yield(a)
        h[k]||= []
        h[k] << a
      end
      return h
    end
    
end


module Math
    
    #無限大。
    INFINITY= 1.0/0.0
    
end


class Array
    
    #selfからランダムにn個選んで返す。
    def shuffle(n= self.size)
      target= self.dup().shuffle!(n)
      return n==self.size ? target : target[0, n]
    end
    
    #selfを破壊的にシャッフルする。
    #nを指定した場合は、selfからランダムに選んだn個をself[0, n]に入れる。
    def shuffle!(n= self.size)
      for i in 0...n
        j= i + rand(self.size-i)
        self[j], self[i]= self[i], self[j]
      end
      return self
    end
    
    #各要素の直積の各要素について繰り返す。
    #e.g. [[1, 2], [3, 4]].enum_for(:each_in_product).to_a() #==> [[1, 3], [1, 4], [2, 3], [2, 4]]
    def each_in_product(prefix= [], &block)
      if self.empty?
        yield(prefix)
      else
        for v in self[0]
          new_prefix= prefix+[v]
          self[1..-1].each_in_product(new_prefix, &block)
        end
      end
    end
    
end


class File
    
    def flock_for_write(&block)
      flock(File::LOCK_EX)
      begin
        yield()
      ensure
        flush()
        flock(File::LOCK_UN)
      end
    end
    
    def flock_for_read(&block)
      flock(File::LOCK_SH)
      begin
        yield()
      ensure
        flock(File::LOCK_UN)
      end
    end
    
end
